import time
import labscript_utils.h5_lock
import h5py
from blacs.tab_base_classes import Worker

class TimeBaseWorker(Worker):

    def init(self):
        global socket; import socket
        timeout = time.time()
        while True: # While loop if connecting does not work the first time.
            if self.mock:
                from unittest import mock
                self.client = mock.Mock()
                # For mock client, specify return value for remote values.
                self.client.recv.return_value = b"Rfreq:39195001|Rampl:236|Rout:1|Rpmon:0|Rpmfr:34|Rpmd:200|Rpmphc:0|Rswpm:0|Rswps:20000000|Rswpp:71000222|Rswpf:700|Rswpt:45000|Rfmon:0|Rfmdev:11|Rplson:0|Rplsfr:66|Rplsdt:50|Rffreq:32000000|Rfampl:156|Ramoffs:0|Rpcbtemp:6175|Rrefstat:0|Rreflev:-85|Rvcclev:2418\r\n"
                break
            try:
                self.client = socket.create_connection((self.hostname,self.port), timeout=2)
            except:
                # If we cannot get connection, wait and try again
                time.sleep(1)
                if time.time()-timeout > 10:
                    raise
                print("No connection, trying again.")
                continue
            break
        print(self.client.recv(200).decode())
        self.smart_cache = {channel:{} for channel in self.channels}

    def front_panel_from_smart_cache(self):
        # Helper function to return only needed parameters to GUI, and convert ampl to dBm
        front_panel = {ch : {
            "gate" : int(self.smart_cache[ch]["Sout"]),
            "freq" : int(self.smart_cache[ch]["Sfreq"]),
            "amp" : 0.1*int(self.smart_cache[ch]["Sampl"]),
        } for ch in self.channels}
        return front_panel
    
    def transition_to_buffered(self, device_name, h5file, initial_values, fresh):
        with h5py.File(h5file,"r") as file:
            group = file[f"devices/{device_name}"]
            for channel in group:
                any_changed = False
                for attr,value in group[channel].attrs.items():
                    if self.smart_cache[channel].get(attr) != group[channel].attrs[attr] or fresh:
                        any_changed = True
                        print(f"Programming {channel}, {attr}={value}")
                        if ":" in attr:
                            # For segment mode
                            self.client.send(f"{channel}|{attr}{value}\r\n".encode())
                        else:
                            self.client.send(f"{channel}|{attr}:{value:.0f}\r\n".encode())
                        self.smart_cache[channel][attr] = group[channel].attrs[attr]
                        time.sleep(0.01) # TODO: What's the best time here? Do we need to wait at all? According to the DIM-3000 manual the rate is 10 commands per second, it seems router can also handle faster times?
                if group[channel].attrs["Sswpm"] != 0 and any_changed:
                    # Set the value of sweep mode (again)
                    # This has to be done after changing any of the sweep parameters.
                    # For simplicity I do that if any parameter was programmed.
                    print(f"Programming {channel}, Sswpm")
                    self.client.send(f"{channel}|Sswpm:{group[channel].attrs['Sswpm']}\r\n".encode())
                    self.smart_cache[channel]["Sswpm"] = {group[channel].attrs['Sswpm']}
                    time.sleep(0.01)
                
                if not initial_values[channel]["gate"]:
                    print(f"Programming {channel}, Sout")
                    self.client.send(f"{channel}|Sout:1\r\n".encode())
                    self.smart_cache[channel]["Sout"] = 1
        return self.front_panel_from_smart_cache()

    def abort_transition_to_buffered(self):
        return True
        
    def abort_buffered(self):
        return True
        
    def transition_to_manual(self):
        return True

    def check_remote_values(self):
        start = time.perf_counter()
        # I rmoved this block, becasue I don't think it's necessary and when we run shots for long this often triggers an error that requires restarts
        # try:
        #     # Call revc in case the TCP server still has some other message queued to send
        #     self.client.recv(1000)
        # except socket.timeout:
        #     pass # Nothing received
        for ch in self.channels:
            self.client.send(ch.encode()+b"|Gpar\r\n")
            time.sleep(0.2)
            try:
                Gpar = self.client.recv(1000).decode() # Rfreq:xxx|Rampl:xxx|Rout:xxx|...
                Gpar = Gpar.split("\n")[-2] # The ADRV-5 also sends Rpcbtemp messages without asking for it. Split off those from the response.
                # Create parameter dictionary from message string
                par = dict(map(lambda str:str.split(":"),Gpar.split("|")[:-1]))
                # Update values in smart cache
                for name,value in par.items():
                    self.smart_cache[ch][f"S{name[1:]}"] = int(value)
            except:
                # Sometimes the TCP servers seems to respond with a wrong answer to Gpar (e.g. b'Rpcbtemp:xxxx\r\n'),
                # maybe because of communication with the http interface or transmission errors.
                # If that happens we just use the saved values and try during the next call again.
                # If that happens during the first check, we need to restart the worker.
                self.logger.debug(f"Exception in check_remote_values of {ch}, returning values of previous check.")
                # try:
                #     self.client.recv(1000)
                # except socket.timeout:
                #     pass # Nothing was in the TCP server's buffer to send
        self.logger.debug(f"Checked remote values ({time.perf_counter()-start:.2f}s)")
        return self.front_panel_from_smart_cache()

    def program_manual(self,values):
        for channel,dds in values.items():
            if dds["freq"] != self.smart_cache[channel]["Sfreq"]:
                self.client.send(f"{channel}|Sfreq:{dds['freq']}\r\n".encode())
                self.smart_cache[channel]["Sfreq"] = dds["freq"]
            if 10*dds["amp"] != self.smart_cache[channel]["Sampl"]:
                dds["amp"] = round(dds["amp"],1)
                self.client.send(f"{channel}|Sampl:{10*dds['amp']}\r\n".encode())
                self.smart_cache[channel]["Sampl"] = 10*dds["amp"]
            if dds["gate"] != self.smart_cache[channel]["Sout"]:
                self.client.send(f"{channel}|Sout:{int(dds['gate'])}\r\n".encode())
                self.smart_cache[channel]["Sout"] = dds["gate"]

    def shutdown(self):
        # for i in self.channels:
        #     ch = f"CH-{i}"
        #     self.client.send(ch.encode() + b"|Sout:0\r\n")
        self.client.shutdown(socket.SHUT_RDWR)
        self.client.close()